import { encoding, toChain, } from "@wormhole-foundation/sdk-base";
import { TokenBridge, deserialize, isTokenTransferDetails, isTransactionIdentifier, isWormholeMessageId, serialize, toNative, toUniversal, } from "@wormhole-foundation/sdk-definitions";
import { signSendWait } from "../common";
import { DEFAULT_TASK_TIMEOUT } from "../config";
import { Wormhole } from "../wormhole";
import { TransferState, hasReachedState, } from "../wormholeTransfer";
export class TokenTransfer {
    wh;
    // state machine tracker
    _state;
    // transfer details
    transfer;
    // txids, populated once transactions are submitted
    txids = [];
    // The corresponding vaa representing the TokenTransfer
    // on the source chain (if its been completed and finalized)
    attestations;
    constructor(wh, transfer) {
        this._state = TransferState.Created;
        this.wh = wh;
        this.transfer = transfer;
    }
    getTransferState() {
        return this._state;
    }
    static async from(wh, from, timeout = 6000) {
        if (isTokenTransferDetails(from)) {
            await TokenTransfer.validateTransferDetails(wh, from);
            const fromChain = wh.getChain(from.from.chain);
            const toChain = wh.getChain(from.to.chain);
            // Apply hackery
            from = {
                ...from,
                ...(await TokenTransfer.destinationOverrides(fromChain, toChain, from)),
            };
            return new TokenTransfer(wh, from);
        }
        let tt;
        if (isWormholeMessageId(from)) {
            tt = await TokenTransfer.fromIdentifier(wh, from, timeout);
        }
        else if (isTransactionIdentifier(from)) {
            tt = await TokenTransfer.fromTransaction(wh, from, timeout);
        }
        else {
            throw new Error("Invalid `from` parameter for TokenTransfer");
        }
        await tt.fetchAttestation(timeout);
        return tt;
    }
    // init from the seq id
    static async fromIdentifier(wh, id, timeout) {
        const vaa = await TokenTransfer.getTransferVaa(wh, id, timeout);
        const automatic = vaa.protocolName === "AutomaticTokenBridge";
        // TODO: the `from.address` here is a lie, but we don't
        // immediately have enough info to get the _correct_ one
        let from = { chain: vaa.emitterChain, address: vaa.emitterAddress };
        let { token, to } = vaa.payload;
        const rescale = (amt, decimals) => decimals > 8 ? amt * 10n ** (decimals - 8n) : amt;
        const amount = rescale(token.amount, await wh.getDecimals(token.chain, token.address));
        let nativeGasAmount = 0n;
        if (automatic) {
            nativeGasAmount = vaa.payload.payload.toNativeTokenAmount;
            from = { chain: vaa.emitterChain, address: vaa.payload.from };
            to = { chain: vaa.payload.to.chain, address: vaa.payload.payload.targetRecipient };
        }
        const details = {
            token: token,
            amount,
            from,
            to,
            automatic,
            nativeGas: nativeGasAmount,
        };
        // TODO: grab at least the init tx from the api
        const tt = new TokenTransfer(wh, details);
        tt.attestations = [{ id: id, attestation: vaa }];
        tt._state = TransferState.Attested;
        return tt;
    }
    static async fromTransaction(wh, from, timeout) {
        const msg = await TokenTransfer.getTransferMessage(wh.getChain(from.chain), from.txid, timeout);
        const tt = await TokenTransfer.fromIdentifier(wh, msg, timeout);
        tt.txids = [from];
        return tt;
    }
    // start the WormholeTransfer by submitting transactions to the source chain
    // returns a transaction hash
    async initiateTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this (eg: state == Created)
            1) get a token transfer transaction for the token bridge given the context
            2) sign it given the signer
            3) submit it to chain
            4) return transaction id
        */
        if (this._state !== TransferState.Created)
            throw new Error("Invalid state transition in `start`");
        const fromChain = this.wh.getChain(this.transfer.from.chain);
        this.txids = await TokenTransfer.transfer(fromChain, this.transfer, signer);
        this._state = TransferState.SourceInitiated;
        return this.txids.map(({ txid }) => txid);
    }
    // wait for the VAA to be ready
    // returns the sequence number
    async fetchAttestation(timeout) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Started)
            1) poll the api on an interval to check if the VAA is available
            2) Once available, pull the VAA and parse it
            3) return seq
        */
        if (this._state < TransferState.SourceInitiated || this._state > TransferState.Attested)
            throw new Error("Invalid state transition in `ready`");
        if (!this.attestations || this.attestations.length === 0) {
            if (this.txids.length === 0)
                throw new Error("No VAAs set and txids available to look them up");
            // TODO: assuming the _last_ transaction in the list will contain the msg id
            const txid = this.txids[this.txids.length - 1];
            const msgId = await TokenTransfer.getTransferMessage(this.wh.getChain(txid.chain), txid.txid, timeout);
            this.attestations = [{ id: msgId }];
        }
        for (const idx in this.attestations) {
            // Check if we already have the VAA
            if (this.attestations[idx].attestation)
                continue;
            this.attestations[idx].attestation = await TokenTransfer.getTransferVaa(this.wh, this.attestations[idx].id, timeout);
        }
        this._state = TransferState.Attested;
        return this.attestations.map((vaa) => vaa.id);
    }
    // finish the WormholeTransfer by submitting transactions to the destination chain
    // returns a transaction hash
    async completeTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Ready)
            1) prepare the transactions and sign them given the signer
            2) submit the VAA and transactions on chain
            3) return txid of submission
        */
        if (this._state < TransferState.Attested)
            throw new Error("Invalid state transition in `finish`. Be sure to call `fetchAttestation`.");
        if (!this.attestations)
            throw new Error("No VAA details available");
        // TODO: when do we get >1?
        const { attestation } = this.attestations[0];
        if (!attestation)
            throw new Error(`No VAA found for ${this.attestations[0].id.sequence}`);
        const toChain = this.wh.getChain(this.transfer.to.chain);
        const redeemTxids = await TokenTransfer.redeem(toChain, attestation, signer);
        this.txids.push(...redeemTxids);
        this._state = TransferState.DestinationInitiated;
        return redeemTxids.map(({ txid }) => txid);
    }
    // Static method to perform the transfer so a custom RPC may be used
    // Note: this assumes the transfer has already been validated with `validateTransfer`
    static async transfer(fromChain, transfer, signer) {
        const senderAddress = toNative(signer.chain(), signer.address());
        const token = transfer.token === "native"
            ? "native"
            : transfer.token.address;
        let xfer;
        if (transfer.automatic) {
            const tb = await fromChain.getAutomaticTokenBridge();
            xfer = tb.transfer(senderAddress, transfer.to, token, transfer.amount, transfer.nativeGas);
        }
        else {
            const tb = await fromChain.getTokenBridge();
            xfer = tb.transfer(senderAddress, transfer.to, token, transfer.amount, transfer.payload);
        }
        return signSendWait(fromChain, xfer, signer);
    }
    // Static method to allow passing a custom RPC
    static async redeem(toChain, vaa, signer) {
        const signerAddress = toNative(signer.chain(), signer.address());
        const xfer = vaa.protocolName === "AutomaticTokenBridge"
            ? (await toChain.getAutomaticTokenBridge()).redeem(signerAddress, vaa)
            : (await toChain.getTokenBridge()).redeem(signerAddress, vaa);
        return signSendWait(toChain, xfer, signer);
    }
    static async isTransferComplete(toChain, vaa) {
        // TODO: converter?
        if (vaa.protocolName === "AutomaticTokenBridge")
            vaa = deserialize("TokenBridge:TransferWithPayload", serialize(vaa));
        const tb = await toChain.getTokenBridge();
        return tb.isTransferCompleted(vaa);
    }
    static async getTransferMessage(chain, txid, timeout) {
        // A Single wormhole message will be returned for a standard token transfer
        const whm = await Wormhole.parseMessageFromTx(chain, txid, timeout);
        if (whm.length !== 1)
            throw new Error("Expected a single Wormhole Message, got: " + whm.length);
        return whm[0];
    }
    static async getTransferVaa(wh, key, timeout) {
        const vaa = typeof key === "string"
            ? await wh.getVaaByTxHash(key, TokenBridge.getTransferDiscriminator(), timeout)
            : await wh.getVaa(key, TokenBridge.getTransferDiscriminator(), timeout);
        if (!vaa)
            throw new Error(`No VAA available after retries exhausted`);
        // TODO: converter
        // Check if its automatic and re-de-serialize
        if (vaa.payloadName === "TransferWithPayload") {
            const { chain, address } = vaa.payload.to;
            const { tokenBridgeRelayer } = wh.config.chains[chain].contracts;
            const relayerAddress = tokenBridgeRelayer ? toUniversal(chain, tokenBridgeRelayer) : null;
            // If the target address is the relayer address, expect its an automatic token bridge vaa
            if (!!relayerAddress && address.equals(relayerAddress)) {
                return deserialize("AutomaticTokenBridge:TransferWithRelay", serialize(vaa));
            }
        }
        return vaa;
    }
    // Lookup the token id for the destination chain given the source chain
    // and token id
    static async lookupDestinationToken(srcChain, dstChain, token) {
        // that will be minted when the transfer is redeemed
        let lookup;
        if (token === "native") {
            // if native, get the wrapped asset id
            lookup = await srcChain.getNativeWrappedTokenId();
        }
        else {
            try {
                const tb = await srcChain.getTokenBridge();
                // otherwise, check to see if it is a wrapped token locally
                lookup = await tb.getOriginalAsset(token.address);
            }
            catch (e) {
                // not a from-chain native wormhole-wrapped one
                lookup = token;
            }
        }
        // if the token id is actually native to the destination, return it
        if (lookup.chain === dstChain.chain) {
            return lookup;
        }
        // otherwise, figure out what the token address representing the wormhole-wrapped token we're transferring
        const dstTb = await dstChain.getTokenBridge();
        const dstAddress = await dstTb.getWrappedAsset(lookup);
        return { chain: dstChain.chain, address: dstAddress };
    }
    static async validateTransferDetails(wh, transfer) {
        if (transfer.payload && transfer.automatic)
            throw new Error("Payload with automatic delivery is not supported");
        if (transfer.nativeGas && !transfer.automatic)
            throw new Error("Gas Dropoff is only supported for automatic transfers");
        const fromChain = wh.getChain(transfer.from.chain);
        const toChain = wh.getChain(transfer.to.chain);
        if (!fromChain.supportsTokenBridge())
            throw new Error(`Token Bridge not supported on ${transfer.from.chain}`);
        if (!toChain.supportsTokenBridge())
            throw new Error(`Token Bridge not supported on ${transfer.to.chain}`);
        if (transfer.automatic && !fromChain.supportsAutomaticTokenBridge())
            throw new Error(`Automatic Token Bridge not supported on ${transfer.from.chain}`);
        if (transfer.amount === 0n)
            throw new Error("Amount cannot be 0");
    }
    static async quoteTransfer(srcChain, dstChain, transfer) {
        const dstToken = await this.lookupDestinationToken(srcChain, dstChain, transfer.token);
        const srcToken = transfer.token === "native" ? await srcChain.getNativeWrappedTokenId() : transfer.token;
        const dstDecimals = await dstChain.getDecimals(dstToken.address);
        const srcDecimals = await srcChain.getDecimals(srcToken.address);
        // truncate the amount for over-the-wire max representation
        const truncate = (amt, decimals) => decimals > 8 ? amt / 10n ** (decimals - 8n) : amt;
        // scale by number of decimals, pair with truncate
        const scale = (amt, decimals) => decimals > 8 ? amt * 10n ** (decimals - 8n) : amt;
        // dedust for source using the number of decimals on the source chain
        // but scale back to original decimals
        const dedust = (amt) => scale(truncate(amt, srcDecimals), srcDecimals);
        // truncate and rescale to fit destination token
        const mintable = (amt) => scale(truncate(amt, srcDecimals), dstDecimals);
        // dedust for transfer over the bridge which truncates to 8 decimals
        let srcAmount = dedust(transfer.amount);
        let dstAmount = mintable(transfer.amount);
        if (!transfer.automatic) {
            return {
                sourceToken: { token: srcToken, amount: srcAmount },
                destinationToken: { token: dstToken, amount: dstAmount },
            };
        }
        // Otherwise automatic
        // If a native gas dropoff is requested, remove that from the amount they'll get
        const _nativeGas = transfer.nativeGas ? mintable(transfer.nativeGas) : 0n;
        dstAmount -= _nativeGas;
        // The fee is also removed from the amount transferred
        // quoted on the source chain
        const stb = await srcChain.getAutomaticTokenBridge();
        const fee = await stb.getRelayerFee(transfer.from.address, transfer.to, srcToken.address);
        dstAmount -= mintable(fee);
        // The expected destination gas can be pulled from the destination token bridge
        let destinationNativeGas = 0n;
        if (transfer.nativeGas) {
            const dtb = await dstChain.getAutomaticTokenBridge();
            destinationNativeGas = await dtb.nativeTokenAmount(dstToken.address, _nativeGas);
        }
        return {
            sourceToken: {
                token: srcToken,
                amount: srcAmount,
            },
            destinationToken: { token: dstToken, amount: dstAmount },
            relayFee: { token: srcToken, amount: fee },
            destinationNativeGas,
        };
    }
    static async destinationOverrides(srcChain, dstChain, transfer) {
        const _transfer = { ...transfer };
        // Bit of (temporary) hackery until solana contracts support being
        // sent a VAA with the primary address
        // Note: Do _not_ override if automatic or if the destination token is native
        // gas token
        if (transfer.to.chain === "Solana" && !_transfer.automatic) {
            const destinationToken = await TokenTransfer.lookupDestinationToken(srcChain, dstChain, _transfer.token);
            // TODO: If the token is native, no need to overwrite the destination address check for native
            //if (!destinationToken.address.equals((await dstChain.getNativeWrappedTokenId()).address))
            _transfer.to = await dstChain.getTokenAccount(_transfer.to.address, destinationToken.address);
        }
        if (_transfer.to.chain === "Sei") {
            if (_transfer.to.chain === "Sei" && _transfer.payload)
                throw new Error("Arbitrary payloads unsupported for Sei");
            // For sei, we reserve the payload for a token transfer through the sei bridge.
            _transfer.payload = encoding.bytes.encode(JSON.stringify({
                basic_recipient: {
                    recipient: encoding.b64.encode(_transfer.to.address.toString()),
                },
            }));
            _transfer.to = Wormhole.chainAddress(_transfer.to.chain, dstChain.config.contracts.translator);
        }
        return _transfer;
    }
    static getReceipt(xfer) {
        const { transfer } = xfer;
        const protocol = transfer.automatic ? "AutomaticTokenBridge" : "TokenBridge";
        const from = transfer.from.chain;
        const to = transfer.to.chain;
        let receipt = {
            protocol,
            request: transfer,
            from: from,
            to: to,
            state: TransferState.Created,
        };
        const originTxs = xfer.txids.filter((txid) => txid.chain === transfer.from.chain);
        if (originTxs.length > 0) {
            receipt = { ...receipt, state: TransferState.SourceInitiated, originTxs: originTxs };
        }
        const att = xfer.attestations && xfer.attestations.length > 0 ? xfer.attestations[0] : undefined;
        const attestation = att && att.id.emitter ? { id: att.id, attestation: att.attestation } : undefined;
        if (attestation && attestation.attestation) {
            receipt = {
                ...receipt,
                state: TransferState.Attested,
                attestation: attestation,
            };
        }
        const destinationTxs = xfer.txids.filter((txid) => txid.chain === transfer.to.chain);
        if (destinationTxs.length > 0) {
            receipt = {
                ...receipt,
                state: TransferState.DestinationInitiated,
                destinationTxs: destinationTxs,
            };
        }
        return receipt;
    }
    // AsyncGenerator fn that produces status updates through an async generator
    // eventually producing a receipt
    // can be called repeatedly so the receipt is updated as it moves through the
    // steps of the transfer
    static async *track(wh, receipt, timeout = DEFAULT_TASK_TIMEOUT, 
    // Optional parameters to override chain context (typically for custom rpc)
    _fromChain, _toChain) {
        const start = Date.now();
        const leftover = (start, max) => Math.max(max - (Date.now() - start), 0);
        _fromChain = _fromChain ?? wh.getChain(receipt.from);
        _toChain = _toChain ?? wh.getChain(receipt.to);
        // Check the source chain for initiation transaction
        // and capture the message id
        if (hasReachedState(receipt, TransferState.SourceInitiated)) {
            if (receipt.originTxs.length === 0)
                throw "Origin transactions required to fetch message id";
            const { txid } = receipt.originTxs[receipt.originTxs.length - 1];
            const msg = await TokenTransfer.getTransferMessage(_fromChain, txid, leftover(start, timeout));
            receipt = { ...receipt, state: TransferState.SourceFinalized, attestation: { id: msg } };
            yield receipt;
        }
        // If the source is finalized, we need to fetch the signed attestation
        // so that we may deliver it to the destination chain
        // or at least track the transfer through its progress
        if (hasReachedState(receipt, TransferState.SourceFinalized)) {
            if (!receipt.attestation.id)
                throw "Attestation id required to fetch attestation";
            const { id } = receipt.attestation;
            const attestation = await TokenTransfer.getTransferVaa(wh, id, leftover(start, timeout));
            receipt = { ...receipt, attestation: { id, attestation }, state: TransferState.Attested };
            yield receipt;
        }
        // First try to grab the tx status from the API
        // Note: this requires a subsequent async step on the backend
        // to have the dest txid populated, so it may be delayed by some time
        if (hasReachedState(receipt, TransferState.Attested) ||
            hasReachedState(receipt, TransferState.SourceFinalized)) {
            if (!receipt.attestation.id)
                throw "Attestation id required to fetch redeem tx";
            const { id } = receipt.attestation;
            const txStatus = await wh.getTransactionStatus(id, leftover(start, timeout));
            if (txStatus && txStatus.globalTx?.destinationTx?.txHash) {
                const { chainId, txHash } = txStatus.globalTx.destinationTx;
                receipt = {
                    ...receipt,
                    destinationTxs: [{ chain: toChain(chainId), txid: txHash }],
                    state: TransferState.DestinationFinalized,
                };
            }
            yield receipt;
        }
        // Fall back to asking the destination chain if this VAA has been redeemed
        // Note: We do not get any destinationTxs with this method
        if (hasReachedState(receipt, TransferState.Attested)) {
            if (!receipt.attestation.attestation)
                throw "Signed Attestation required to check for redeem";
            receipt = {
                ...receipt,
                state: (await TokenTransfer.isTransferComplete(_toChain, receipt.attestation.attestation))
                    ? TransferState.DestinationFinalized
                    : TransferState.Attested,
            };
            yield receipt;
        }
        yield receipt;
    }
}
//# sourceMappingURL=tokenTransfer.js.map